/*
 * Copyright (C) 2023  Valentin Lorentz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License version 3,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

use crate::widgets::Prerender;
use futures::stream::FuturesUnordered;
use futures::StreamExt;
use matrix_sdk::async_trait;
use nonempty::NonEmpty;
use ratatui::text::Text;

mod log;
pub use log::LogBuffer;
mod room;
pub use room::RoomBuffer;

pub struct BufferItem<'buf> {
  pub text: Text<'buf>,
  pub prerender: &'buf Prerender,
}

#[async_trait]
pub trait Buffer: Send + Sync {
  /// A short human-readable name for the room, eg. to show in compact buflist
  fn short_name(&self) -> String;
  /// If this is a room buffer, returns the associated room id.
  fn room_id(&self) -> Option<&matrix_sdk::ruma::RoomId> {
    None
  }
  /// Returns if there are any updates to apply.
  async fn poll_updates(&mut self);
  fn content<'a>(&'a self) -> Box<dyn Iterator<Item = BufferItem<'a>> + 'a>;
  /// Called when the user is being showned the oldest items this buffer returned.
  ///
  /// This should return immediately, not waiting for anything to be loaded.
  fn request_back_pagination(&self, num: u16) {}
}

pub struct Buffers {
  buffers: NonEmpty<Box<dyn Buffer>>,
  active_index: usize,
}

impl Buffers {
  pub fn new(initial_buffer: Box<dyn Buffer>) -> Self {
    Self {
      buffers: NonEmpty::new(initial_buffer),
      active_index: 0,
    }
  }

  pub async fn poll_updates(&mut self) {
    self
      .iter_mut()
      .map(|buf| buf.poll_updates())
      .collect::<FuturesUnordered<_>>()
      .next()
      .await
      .expect("poll_updates reached the end of the never-ending stream");
  }

  pub fn iter(&self) -> impl Iterator<Item = &dyn Buffer> {
    self.buffers.iter().map(|buffer_box| &**buffer_box)
  }

  pub fn iter_mut(&mut self) -> impl Iterator<Item = &mut Box<dyn Buffer>> {
    self.buffers.iter_mut()
  }

  pub fn push(&mut self, buffer: Box<dyn Buffer>) {
    self.buffers.push(buffer)
  }

  pub fn active_index(&self) -> usize {
    self.active_index
  }

  pub fn set_active_index(&mut self, index: isize) {
    self.active_index = index.wrapping_rem_euclid(self.buffers.len() as isize) as usize;
  }

  pub fn active_buffer(&self) -> &dyn Buffer {
    &**self
      .buffers
      .get(self.active_index)
      .expect("Active buffer index does not exist")
  }

  pub fn active_buffer_mut(&mut self) -> &mut Box<dyn Buffer> {
    self
      .buffers
      .get_mut(self.active_index)
      .expect("Active buffer index does not exist")
  }
}
