/*
 * Copyright (C) 2023  Valentin Lorentz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License version 3,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

use std::sync::atomic::{AtomicU16, Ordering};
use std::sync::{Arc, OnceLock};

use color_eyre::eyre::{eyre, Result};
use eyeball_im::VectorDiff;
use futures::{FutureExt, Stream, StreamExt};
use itertools::Itertools;
use matrix_sdk::async_trait;
use matrix_sdk::ruma::{OwnedRoomId, RoomId};
use matrix_sdk::Client;
use matrix_sdk::Room;
use matrix_sdk_ui::timeline::{
  BackPaginationStatus, PaginationOptions, RoomExt, Timeline, TimelineItem, TimelineItemKind,
  VirtualTimelineItem,
};
use ratatui::text::Text;
use smallvec::SmallVec;

use super::{Buffer, BufferItem};
use crate::widgets::Prerender;

pub struct SingleClientRoomBuffer {
  room_id: OwnedRoomId,
  client: Client,
  items: imbl::vector::Vector<(String, Prerender)>,
  // TODO: get rid of this trait object, we know it's matrix_sdk_ui::timeline::TimelineStream
  stream: Box<dyn Stream<Item = Vec<VectorDiff<Arc<TimelineItem>>>> + Send + Sync + Unpin>,
  timeline: Arc<Timeline>,
  back_pagination_request: AtomicU16,
}

impl SingleClientRoomBuffer {
  async fn poll_updates(&mut self) {
    let back_pagination_request = self.back_pagination_request.swap(0, Ordering::Relaxed);
    if back_pagination_request > 0 {
      // TODO: run this concurrently with stream.next() below
      self.spawn_back_pagination(back_pagination_request).await;
    }
    if let Some(changes) = self.stream.next().await {
      for change in changes {
        change
          .map(|item| (self.format_timeline_item(item), Prerender::new()))
          .apply(&mut self.items);
      }
    }
  }

  fn format_timeline_item(&self, tl_item: impl AsRef<TimelineItem>) -> String {
    match tl_item.as_ref().kind() {
      TimelineItemKind::Event(event) => {
        use matrix_sdk_ui::timeline::TimelineItemContent::*;
        let sender = event
          .sender()
          .as_str()
          .strip_prefix("@")
          .expect("missing @ prefix");
        match event.content() {
          Message(message) => format!("   <{}> {}", sender, message.body().replace("\n", "\n    ")),
          RedactedMessage => format!("xx <{}> [redacted]", sender),
          Sticker(sticker) => format!("st <{}> {}", sender, sticker.content().body),
          UnableToDecrypt(_) => format!("xx <{}> [unable to decrypt]", sender),
          MembershipChange(change) => {
            use matrix_sdk_ui::timeline::MembershipChange::*;
            if change.user_id() == event.sender() {
              let Some(change_kind) = change.change() else {
                return format!("--- {} made incomprehensible changes to themselves", sender);
              };
              match change_kind {
                None => format!("--- {} made no discernable changes to themselves", sender),
                Error => format!(
                  "xxx {} made a change to themselves that made matrix-sdk-ui error",
                  sender
                ),
                Joined => format!("--> {} joined", sender),
                Left => format!("<-- {} left", sender),
                Banned => format!("-x- {} banned themselves", sender),
                Unbanned => format!("-x- {} unbanned themselves", sender),
                Kicked => format!("<!- {} kicked themselves", sender),
                Invited => format!("-o- {} invited themselves", sender),
                KickedAndBanned => format!("<!x {} kicked and banned themselves", sender),
                InvitationAccepted => format!("-o> {} accepted an invite", sender),
                InvitationRejected => format!("-ox {} rejected an invite", sender),
                InvitationRevoked => format!("--x {} revoked an invite", sender),
                Knocked => format!("-?> {} knocked", sender),
                KnockAccepted => format!("-?o {} accepted a knock", sender),
                KnockRetracted => format!("-?x {} retracted a knock", sender),
                KnockDenied => format!("-?x {} denied a knock", sender),
                NotImplemented => format!(
                  "xxx {} made a change matrix-sdk-ui does not support yet",
                  sender
                ),
              }
            } else if change.user_id() == "" {
              let Some(change_kind) = change.change() else {
                return format!("--- {} made incomprehensible changes", sender);
              };
              match change_kind {
                None => format!("--- {} made no discernable changes", sender),
                Error => format!("xxx {} made a change that made matrix-sdk-ui error", sender),
                Joined | Left | Banned | Unbanned | Kicked | Invited | KickedAndBanned
                | InvitationAccepted | InvitationRejected | InvitationRevoked | Knocked
                | KnockAccepted | KnockRetracted | KnockDenied => {
                  format!("--> {} made a non-sensical change: {:?}", sender, change)
                },
                NotImplemented => format!(
                  "xxx {} made a change matrix-sdk-ui does not support yet",
                  sender
                ),
              }
            } else {
              let target = change
                .user_id()
                .as_str()
                .strip_prefix("@")
                .expect("missing @ prefix");
              let Some(change_kind) = change.change() else {
                return format!("--- {} made incomprehensible changes to {}", sender, target);
              };
              match change_kind {
                None => format!("--- {} made no discernable changes to {}", sender, target),
                Error => format!(
                  "xxx {} made a change to {} that made matrix-sdk-ui error",
                  sender, target
                ),
                Joined | Left => format!(
                  "--> {} made a non-sensical change to {}: {:?}",
                  sender, target, change
                ),
                Banned => format!("-x- {} banned {}", sender, target),
                Unbanned => format!("-x- {} unbanned {}", sender, target),
                Kicked => format!("<!- {} kicked {}", sender, target),
                Invited => format!("-o- {} invited {}", sender, target),
                KickedAndBanned => format!("<!x {} kicked and banned {}", sender, target),
                InvitationAccepted => format!("-o> {} accepted an invite to {}", sender, target),
                InvitationRejected => format!("-ox {} rejected an invite to {}", sender, target),
                InvitationRevoked => format!("--x {} revoked an invite to {}", sender, target),
                Knocked => format!("-?> {} made {} knock", sender, target),
                KnockAccepted => format!("-?o {} accepted {}'s knock", sender, target),
                KnockRetracted => format!("-?x {} retracted {}'s knock", sender, target),
                KnockDenied => format!("-?x {} denied {}'s knock", sender, target),
                NotImplemented => format!(
                  "xxx {} made a change to {} that matrix-sdk-ui does not support yet",
                  sender, target
                ),
              }
            }
          },

          ProfileChange(_) => format!("--- {} updated their profile", sender),
          OtherState(state) => {
            if state.state_key() == "" {
              format!("--- {} changed the room: {:?}", sender, state.content())
            } else {
              format!(
                "--- {} changed {}: {:?}",
                sender,
                state.state_key(),
                state.content()
              )
            }
          },
          FailedToParseMessageLike { event_type, error } => format!(
            "xxx {} sent a {} message that made matrix-sdk-ui error: {:?}",
            sender, event_type, error
          ),
          FailedToParseState {
            event_type,
            state_key,
            error,
          } => {
            format!(
              "xxx {} made a {} change to {} that made matrix-sdk-ui error: {:?}",
              sender, event_type, state_key, error
            )
          },
          Poll(_) => format!("-?- {} acted on a poll", sender),
        }
      },
      TimelineItemKind::Virtual(VirtualTimelineItem::ReadMarker) => {
        format!("---- read marker ----")
      },
      TimelineItemKind::Virtual(VirtualTimelineItem::DayDivider(day_divider)) => {
        format!("---- day divider: {:?} ----", day_divider)
      },
    }
  }

  async fn spawn_back_pagination(&self, num: u16) {
    let room_id = self.room_id.clone();
    let timeline = self.timeline.clone();
    let mut back_pagination_status = timeline.back_pagination_status();
    if back_pagination_status.get() == BackPaginationStatus::Paginating {
      // We are already waiting for a backfill from the server
      return;
    }
    tokio::spawn(async move {
      tracing::info!("Starting pagination for {}", room_id);
      timeline
        .paginate_backwards(matrix_sdk_ui::timeline::PaginationOptions::until_num_items(
          num, num,
        ))
        .await
        .unwrap_or_else(|e| tracing::error!("Failed to paginate {} backward: {}", room_id, e));
      tracing::info!("Ended pagination for {}", room_id);
    });

    // Wait for the task we just spawned to change the status, so we don't risk starting
    // a new one in the meantime
    back_pagination_status.next().await;
  }
}

pub struct RoomBuffer {
  room_id: OwnedRoomId,

  // It's unlikely users will join the same room with more than one account;
  // avoid a useless heap allocation for the usual case.
  buffers: SmallVec<[SingleClientRoomBuffer; 1]>,
}

fn f(b: RoomBuffer) {
  g(b);
}
fn g<B: Send>(b: B) {}

impl RoomBuffer {
  pub async fn new(initial_client: Client, room_id: OwnedRoomId) -> Result<Self> {
    let mut self_ = RoomBuffer {
      buffers: SmallVec::new(),
      room_id,
    };
    self_.add_client(initial_client).await?;
    Ok(self_)
  }

  pub async fn add_client(&mut self, client: Client) -> Result<()> {
    let timeline = client
      .get_room(&self.room_id)
      .ok_or_else(|| {
        tracing::error!(
          "Adding {:?} for {:?}, but it does not know this room ({} other clients know this room)",
          client,
          self.room_id,
          self.buffers.len()
        );
        eyre!("Unknown room {} for client {:?}", self.room_id, client)
      })?
      .timeline_builder()
      .build()
      .await;
    let (items, stream) = timeline.subscribe_batched().await;
    tracing::info!(
      "Added client for {}, initial items: {:?}",
      self.room_id,
      items
    );
    self.buffers.push(SingleClientRoomBuffer {
      room_id: self.room_id.clone(),
      client,
      timeline: Arc::new(timeline),
      items: items // FIXME: it's always empty. why?
        .into_iter()
        .map(|item| (format!("Initial item: {:#?}", item), Prerender::new()))
        .collect(),
      stream: Box::new(stream),
      back_pagination_request: AtomicU16::new(0),
    });
    Ok(())
  }
}

#[async_trait]
impl Buffer for RoomBuffer {
  fn short_name(&self) -> String {
    self.room_id.as_str().to_owned()
  }

  fn room_id(&self) -> Option<&RoomId> {
    Some(&self.room_id)
  }

  async fn poll_updates(&mut self) {
    futures::future::join_all(
      self
        .buffers
        .iter_mut()
        .map(SingleClientRoomBuffer::poll_updates),
    )
    .await;
  }

  fn content<'a>(&'a self) -> Box<dyn Iterator<Item = BufferItem<'a>> + 'a> {
    // TODO: merge buffers, etc.
    Box::new(
      self
        .buffers
        .first()
        .unwrap_or_else(|| panic!("No sub-buffer for {}", self.room_id))
        .items
        .iter()
        .rev()
        .map(|(line, prerender)| BufferItem {
          text: Text::raw(line),
          prerender,
        }),
    )
  }

  fn request_back_pagination(&self, num: u16) {
    // TODO: pick a client at random instead of just the first one, etc.
    self
      .buffers
      .first()
      .unwrap_or_else(|| panic!("No sub-buffer for {}", self.room_id))
      .back_pagination_request
      .fetch_max(num, Ordering::Relaxed);
  }
}
