/*
 * Copyright (C) 2023  Valentin Lorentz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License version 3,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

use std::ops::DerefMut;

use color_eyre::eyre::{Result, WrapErr};
use crossterm::event::{KeyCode, KeyEvent, KeyEventKind, KeyModifiers};
use ratatui::{prelude::*, widgets::*};

use crate::components::Action;
use crate::widgets::prerender::{PrerenderInner, PrerenderValue};
use crate::widgets::{BottomAlignedParagraph, OverlappableWidget};

use super::Component;
use crate::buffers::BufferItem;

#[derive(Default)]
pub struct Backlog {
  scroll: u64,
}

impl Component for Backlog {
  fn handle_key_events(&mut self, key: KeyEvent) -> Result<Option<Action>> {
    match key {
      KeyEvent {
        code: KeyCode::PageUp,
        modifiers: KeyModifiers::NONE,
        kind: KeyEventKind::Press,
        state: _,
      } => {
        self.scroll_up(20); // TODO: use the component height
      },
      KeyEvent {
        code: KeyCode::PageDown,
        modifiers: KeyModifiers::NONE,
        kind: KeyEventKind::Press,
        state: _,
      } => {
        self.scroll_down(20) // TODO: use the component height
      },
      _ => {},
    }

    Ok(None)
  }

  fn draw(
    &mut self,
    frame: &mut Frame<'_>,
    area: Rect,
    buffers: &crate::buffers::Buffers,
  ) -> Result<()> {
    let active_buffer = buffers.active_buffer();
    let items = active_buffer.content();
    let unused_top_lines = self.draw_items(frame.buffer_mut(), area, items)?;

    // There is empty room on screen, ask the buffer to fetch more backlog if it can
    if unused_top_lines > 0 {
      active_buffer.request_back_pagination(100);
    }

    Ok(())
  }
}

impl Backlog {
  pub fn scroll_up(&mut self, lines: u64) {
    self.scroll = self.scroll.saturating_add(lines);
  }
  pub fn scroll_down(&mut self, lines: u64) {
    self.scroll = self.scroll.saturating_sub(20);
  }
  /// Returns how many lines are unused at the top of the area.
  pub fn draw_items<'a>(
    &mut self,
    frame_buffer: &mut Buffer,
    area: Rect,
    mut items: impl Iterator<Item = BufferItem<'a>>,
  ) -> Result<u16> {
    let block = Block::new().borders(Borders::ALL);
    let mut text_area = block.inner(area);
    block.render(area, frame_buffer);

    let mut scroll = self.scroll;

    // Skip widgets at the bottom (if scrolled up), and render the first visible one
    loop {
      let Some(item) = items.next() else {
        break;
      };
      let expected_height = match item.prerender.0.lock().unwrap().deref_mut() {
        Some(PrerenderInner {
          key,
          value: PrerenderValue::Rendered(buf),
        }) if *key == text_area.width => buf.area().height as u64,
        Some(PrerenderInner {
          key,
          value: PrerenderValue::NotRendered(height),
        }) if *key == text_area.width => *height,
        prerender => {
          let widget = BottomAlignedParagraph::new(item.text.clone());
          let expected_height = widget.height(text_area.width);
          *prerender = Some(PrerenderInner {
            key: text_area.width,
            value: PrerenderValue::NotRendered(expected_height),
          });
          expected_height
        },
      };

      if scroll.saturating_sub(expected_height) > text_area.height.into() {
        // Paragraph is too far down, not displayed
        scroll -= expected_height;
        continue;
      }

      // TODO: cache this
      let widget = BottomAlignedParagraph::new(item.text).scroll(scroll);
      let height = widget.render_overlap(text_area, frame_buffer);
      text_area.height = text_area.height.saturating_sub(height);

      scroll = scroll.saturating_sub(expected_height);
      if scroll == 0 {
        break;
      }
    }
    if text_area.height == 0 {
      // No more room to display other paragraphs, stop now
      return Ok(0);
    }

    // Render other widgets
    for item in items {
      let height = match item.prerender.0.lock().unwrap().deref_mut() {
        Some(PrerenderInner {
          key,
          value: PrerenderValue::Rendered(buf),
        }) if *key == text_area.width => {
          // We already rendered it, copy the buffer.
          assert_eq!(text_area.width, buf.area.width);
          let top_padding = text_area.height.saturating_sub(buf.area.height);
          let skip_top_lines = buf.area.height.saturating_sub(text_area.height);
          copy_buffer(
            buf,
            Rect {
              x: 0,
              y: 0,
              width: text_area.width,
              height: buf.area.height - skip_top_lines,
            },
            frame_buffer,
            Rect {
              y: text_area.y + top_padding,
              height: text_area.height - top_padding,
              ..text_area
            },
          );

          buf.area().height
        },
        prerender => {
          let widget = BottomAlignedParagraph::new(item.text);
          let height = widget.render_overlap(text_area, frame_buffer);

          // If the whole widget fits in the text_area, copy the drawn result to a buffer
          // for caching
          if height <= text_area.height {
            let mut buf = ratatui::buffer::Buffer::empty(Rect {
              x: 0,
              y: 0,
              width: text_area.width,
              height,
            });
            let top_padding = text_area.height.saturating_sub(buf.area.height);
            copy_buffer(
              frame_buffer,
              Rect {
                y: text_area.y + top_padding,
                height: text_area.height - top_padding,
                ..text_area
              },
              &mut buf,
              Rect {
                x: 0,
                y: 0,
                width: text_area.width, // TODO: only copy the width actually drawn by the widget
                height: height,
              },
            );
            *prerender = Some(PrerenderInner {
              key: text_area.width,
              value: PrerenderValue::Rendered(buf),
            });
          }

          height
        },
      };

      assert!(area.height >= height, "{:?} {}", area, height);
      text_area.height = text_area.height.saturating_sub(height); // Remove lines at the bottom used by this paragraph
      if text_area.height == 0 {
        // No more room to display other paragraphs, stop now
        return Ok(0);
      }
    }

    Ok(text_area.height)
  }
}

fn copy_buffer(src: &Buffer, src_area: Rect, dst: &mut Buffer, dst_area: Rect) {
  assert_eq!(src_area.width, dst_area.width, "width mismatch");
  assert_eq!(src_area.height, dst_area.height, "height mismatch");

  assert!(
    src_area.x + src_area.width <= src.area.width,
    "src width overflow"
  );
  assert!(
    src_area.y + src_area.height <= src.area.height,
    "src height overflow"
  );
  assert!(
    dst_area.x + dst_area.width <= dst.area.width,
    "dst width overflow"
  );
  assert!(
    dst_area.y + dst_area.height <= dst.area.height,
    "dst height overflow"
  );

  for (y, line) in src
    .content()
    .chunks(src.area.width as usize)
    .skip(src_area.y as usize)
    .take(dst_area.height as usize)
    .enumerate()
  {
    for (x, cell) in line
      .into_iter()
      .skip(src_area.x as usize)
      .take(src_area.width as usize)
      .enumerate()
    {
      *dst.get_mut(dst_area.x + (x as u16), dst_area.y + (y as u16)) = src
        .get(src_area.x + (x as u16), src_area.y + (y as u16))
        .clone();
    }
  }
}
