/*
 * Copyright (C) 2023  Valentin Lorentz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License version 3,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

use color_eyre::eyre::{Result, WrapErr};
use ratatui::{prelude::*, widgets::*};
use tokio::sync::mpsc::UnboundedSender;
use tokio::sync::OnceCell;

use super::Component;
use crate::{
  action::Action,
  config::{Config, KeyBindings},
};

#[derive(Default)]
pub struct Buflist {
  command_tx: Option<UnboundedSender<Action>>,
  config: OnceCell<Config>,
}
impl Buflist {
  pub fn new() -> Self {
    Self::default()
  }
}

impl Component for Buflist {
  fn register_action_handler(&mut self, tx: UnboundedSender<Action>) -> Result<()> {
    self.command_tx = Some(tx);
    Ok(())
  }

  fn register_config_handler(&mut self, config: Config) -> Result<()> {
    self
      .config
      .set(config)
      .context("Buflist config was already set")
  }

  fn update(&mut self, action: &Action) -> Result<Option<Action>> {
    match action {
      Action::Tick => {},
      _ => {},
    }
    Ok(None)
  }

  fn draw(
    &mut self,
    frame: &mut Frame<'_>,
    area: Rect,
    buffers: &crate::buffers::Buffers,
  ) -> Result<()> {
    frame.render_widget(
      Paragraph::new(
        buffers
          .iter()
          .enumerate()
          .map(|(i, buf)| {
            let mut style = Style::default();
            if i == buffers.active_index() {
              style = style.bold();
            }
            Span::styled(buf.short_name(), style).into()
          })
          .collect::<Vec<Line<'_>>>(),
      )
      .block(Block::new().borders(Borders::ALL)),
      area,
    );

    Ok(())
  }
}
